%+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++   
%
% Part of softare library HSpec ("Hilbert Spectrum")
% Free software written by  Sofia Olhede and Andrew Walden,
% Dept. of Mathematics, Imperial College, London SW7 2BZ, UK
% November 2004
%
% For terms of use and warranty disclaimer please refer to the
% README file
%
%++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++
%
function [W,E,S] = modwptgenr(X, kfilt, J, Graph)
%
% Authors: Bill Constantine and Andrew Walden
%
%  [W,E,S] = MODWPT(X, kfilt, J, Graph)
%
%  Performs the Maximum Overlap Discrete Wavelet Packet Transform 
%  for J {Full} scales
%  using Fejer Korovkin  filters (see below) on the 1D input signal X.  
%
% kfilt==0 : uses FejerKorovkin filter of length 18, FK(18)
% kfilt==1 : FK(22)
%
% W is a 2D array whose elements are W_(n,t) where n is the global node index 
% and t is the "time" index of the resulting basis.
%
%  So the matrix W has rows of the following form:
%
%  global node
%  index
%  1                 X
%  2                 W_{1,0,t}  (j=1,n=0)
%  3                 W_{1,1,t}  (j=1,n=1)
%  4                 W_{2,0,t}  (j=2,n=0)
%  5                 W_{2,1,t}  (j=2,n=1)
%                      ..
%  7                 W_{2,3,t}
%  8                 W_{3,0,t}
%                       ..
%  15                 W_{3,7,t}  etc
%
% If E or S is included as a return argument, the energy
% or Shannon-Weaver entropy is returned in vector E or S,
% respectively, where each element corresponds to a node
% in the wavelet packet tree numbered from 1 (the original signal)
% to M=sum(2^j)+1 for j = 1:J. Note: the original signal is not 
% normalized in any way. Any normalization should be done prior to
% calling this routine. If Graph{0} > 0 a graph of the modwpt
% coefficients
% is produced such that at each node in the decomposition, the
% resulting
% basis is plotted.
maxk=2; % max number of filters defined
if (nargin < 1)
   disp('Error:  Must supply input X.'); break;
else
   lx = length(X);
   N = floor(lx/2)*2;
   Kfull =  floor(log(N)/log(2));
end
if (nargin < 4) Graph = 0; end;
if (nargin < 3) J = Kfull; end;
if (nargin < 2) 
        disp('Error: Must supply type of filter required');
return;
end;
if (length(J) < 1) J = Kfull; end;
%if (length(Bases) < 1) Bases = 1; end;

FiltStr=str2mat('FK(18)','FK(22)');
%
X=transpose(X(:));
X = X(1:N);             % Make X even in length
if ( N < lx) disp(['Input shortened to ' int2str(N) ' points from' ...
	int2str(lx) ' points.']); end;
%
if or(kfilt<0,kfilt>maxk)
disp('Choose filters in approproate range')
break;break;
else
G = wavgfiltsr(kfilt);
end;
disp(['Filter:  ' deblank(FiltStr(kfilt+1,:)) ' loaded'])

%
H=flipud(G).*(-1).^(0:length(G)-1)';
L = length(H);
%  Rescale wavelet filters for MODWPT
fac1 = sqrt(2);H = H/fac1;G = G/fac1;

H=transpose(H(:));
G=transpose(G(:));  % Make into row vectors

% W's indices are W_(j,n,t)
nnodes = 2^(J+1)-1;
W = zeros(nnodes,N);
WW = zeros(1,N);
WWW = zeros(1,N);

if nargout > 1, 
   nnodes = sum(2.^(0:J)); % Total nodes including original sequence X
   E_t = zeros(1,N);
   E = zeros(1,nnodes);
   E_t = X.^2;
   E(1) = sum(E_t);
   ie = 2;
end

if nargout > 2, 
   S = zeros(1,nnodes);
   if any(E_t == 0), E_t = E_t(E_t > 0); end;
   S(1) = -sum(E_t.*log(E_t));
   is = 2;
end

W(1,:)=X;
FilterIndexRange = 0:L-1;
fprintf('Calculating scale j = ');

b=zeros(L,N); % W time index matrix 
base=1;       % Uses base 1 while the pseudocode is in base 0       
% j is still base 0, we are just calculating from scales j = 1 on up
for j = 1:J,   % Level loop
   fprintf('%i',j);
   if j < J, fprintf(',');end;
   for t=0:N-1,
% Form W time index matrix
		b(:,t+base)=(mod(t-2^(j-1)*FilterIndexRange,N)+base)';
	end;  
   
   for n = 0:(2^j)-1,
% Bin loop (bins numbered 0...(2^j)-1 for each level j
      
      if any(mod(n,4) == [0 3]) u = G; else u = H; end;
% Grab the correct node to filter      
      nn=floor(n/2); 	  
      WW = W(2^(j-1)+nn,:);                     
      WWW = u*WW(b); % [1xL]*[LxN]=[1xN]
      nodeindex = 2^j+n;  %W_j,n,t -> W'_(2^(j)+n),t for j,n in base 0
      W(nodeindex,:) = WWW;     
      
      if nargout > 1, 
         E_t = WWW.^2; 
         E(ie) = sum(E_t);
         ie = ie + 1;
      end
      
      if nargout > 2,
         if any(E_t == 0), E_t = E_t(E_t > 0); end;
         S(is) = -sum(E_t.*log(E_t));
         is = is + 1;
      end
      
   end
end

if Graph,
   fprintf('\nPlotting results ...');
   dwptplot(W,1)
   fprintf('Done.\n')
end

if nargout < 1, W =[]; 
end;

fprintf('\n');

